#!/usr/bin/python
# -*- coding: utf-8 -*-

"""Python implementation of the TCLserv protocol

TCLserv protocol is used to supervise Pocolibs-based module 
(like LAAS' Genom modules) through TCLserv.

pypoco provides the PocoServProxy class that allows to connect to a remote
TCLserv server. Once connected, it is possible to load/unload modules on the
remote end, execute synchronous or asynchronous request (by providing a 
callback function) and other management operation (like listing available 
modules, listing available service for a module, aborting a pending request,
etc.)

When loading a module, pypoco creates a new PocoModule instance. Services 
(ie, requests) provided by the module are automatically added as regular 
Python method, and can be directly invoked.

Lastly, a helper function, pypoco.discover(), allows to automatically create
PocoServProxy instances and PocoModule for all available module.
This method is simply called with a list of hostnames (and optionally ports,
 if not using the default 9472 port).

Example of an interactive session:

> import pypoco
> servers, modules = pypoco.discover(['jido-base', 'jido-top'])
Loaded modules pom, viam on 2 server(s)

#Using modules:
> pom = modules['pom']
> pom.SetPos('test', 1, 1, 1, 0, 0, 0)

#Loading new modules on a remote host:
> jidotop = servers['jido-top']
> viman = jidotop.load('viman')
> viman.GetObjectPose(1)



TCLserv protocol:
1/ send "HELLO\n" on the outbound socket
2/ tclserv answers "HELLO [id]\r\n" on the same socket. 'id' identify the 
client during the session.

All command ends with a line break.

Then, you have two kind of interations:
- RPC-style call like "RQST [request]". In these cases, TCL serv answers 
with either "OK [rqst_id]" or "ERROR [n] [err_msg]" (where n seems to be
 ;-) the error number returned by modules or 0 for TCL serv internal 
 errors or 1 for other errors.

- messages from TCL serv (either ACK when a module acknowledge it has 
received the request or TERM when a request finishes):
  -> "[rqst_id] [cmd name] ACK [activity_id]"
  -> "[rqst_id] [cmd name] TERM [rqst_answer]"
    rqst_answer can be either "OK [result]" or an error message from 
    the module.
  
List of RPC commands:
 - HELLO|BYE: initilizes/terminates a connection to the server, and answer
the socket id.
 - REPLYTO [socket_id]: allows to specify the socket for server's
answers. To use only one socket for inbound and outbound communication, 
use the socket id returned by HELLO.
 - LM|UNLM: load or unload a module in TCL serv
 - RQST [request]: send a request
 - ABORT [activity_id]: try to abort the given activity
 - KILL [module_name] abort|verbose|showtime|resume: send the corresponding
 signal to a loaded module.
 - DIE: kill tclserv

Then any valid TCL code can be send and will be executed.
It noteworthy includes:
 - "exec -- [cmd]&" to spawn a new process, including new modules
 - "cs::lsmbox" to list all H2 mailboxes present on the server
"""

__author__ = "Séverin Lemaignan <slemaign@laas.fr>"
__date__ = "17 Aug 2010"
__version__ = "1.0"

import time
import logging
import socket
import select
from threading import Thread, Lock
try:
	from Queue import Queue
except ImportError:
	from queue import Queue #Python 3.0

class PocoRemoteError(Exception):
	def __init__(self, value):
		self.value = value
	def __str__(self):
		return repr(self.value)

class PocoModule():
	"""PocoModules are created when a module is loaded on TCLserv and 
	expose all available services as regular Python methods.
	"""
	def __init__(self, name, serv):
		self.name = name
		self._serv = serv
		
		for service in serv.list_services(name):
			self.add_service(service)

	def add_service(self, m):
		def innermethod(*args):
			callback = None
			wait_for_ack=True
			
			#Check if the first argument is callable. If true, assume it's a
			#callback, and do the request in non-blocking mode.
			if not len(args) == 0 and hasattr(args[0], '__call__'):
				callback = args[0]
				args = args[1:]
				
			req = "::" + self.name + "::" + m + " " + " ".join([str(a) for a in args])
			logging.log(4, "Sending \"" + req + "\" to " + m + " module.")
			return self._serv.request(req, callback, wait_for_ack)
				
		innermethod.__doc__ = "This method is a proxy for the %s %s service." % (self.name, m) + \
		"If the first argument is a callable (like a function), we assume it's " + \
		"a callback and the request is sent in non-blocking mode. Else, we execute " + \
		"the request in blocking mode, ie, we wait for the request to complete and " + \
		"return the module response."
		innermethod.__name__ = m
		setattr(self,innermethod.__name__,innermethod)
	
class PocoRequest():
	def __init__(self, rqst_id):
		self.rqst_id = rqst_id
		self.state = 'INITIAL'
		self.status = None #OK or FAILED
		self.activity_id = None
		self.answer = None
		self.callback = None
		
class PocoServProxy(Thread):
	"""This class acts as a proxy between Python application and TCLserv.
	"""
	def __init__(self, host_port = ('localhost', 9472)):
		"""The constructor expects a tuple (hostname, port) as parameter.
		"""
		Thread.__init__(self)
		
		self._ack_lock = Lock()
		
		self._poco_responses_queue = Queue()
		
		self._running = True
		
		#The TCL serv socket
		self._tclserv = None
		
		#Our ID in the TCL serv
		self._id = ''
		
		#The dict of loaded modules, as PocoModule objects
		self._loaded_modules = {}
		
		"""This map stores the ids of currently pending requests and the corresponding
		Request instance.
		"""
		self.requests = {}
		
		try:
			#create an INET, STREAMing socket
			self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			
			self.s.connect(host_port)
			self._tclserv = self.s.makefile()
		except socket.error:
			self.s.close()
			raise PocoRemoteError('Unable to connect to the remote TCL server on ' + \
								 host_port[0] + '. ' + \
								 'Check it is running and that you provided the' + \
								 ' right host and port.')
		
		self._tclserv.write("HELLO\r\n")
		self._tclserv.flush()
		
		valid, self._id = self._tclserv.readline().splitlines()[0].split()
		
		if valid != 'HELLO':
			raise PocoRemoteError('Could connect to the TCL serv, but initial ' + \
				'handshake failed')
		
		#Set the REPLY socket
		self._tclserv.write("REPLYTO %s\r\n" % self._id)
		self._tclserv.flush()
		
		valid, id = self._tclserv.readline().splitlines()[0].split()
		if valid != 'OK' and self._id != id:
			raise PocoRemoteError('Error when setting up the reply channel')
		
		logging.info("Known as " + self._id + " by the TCL server")

		self.start()
	
	def run(self):
		""" This method reads and writes to/from the TCL server.
		When a new request is pushed in _poco_requests_queue, it is send to the 
		server. When the server answer smthg, we store it in the _poco_requests_queue
		"""
		
		logging.debug("Launching socket reading thread.")
		
		while self._running:
			raw = self._read_tclserv_answer()
			
			if not raw:
				continue
			
			logging.debug("<<< " + str(raw))
			
			#If answer to a RPC call, push it forward, outside of this thread.
			if raw[0] in ['OK', 'ERROR']:
					self._poco_responses_queue.put(raw)
					continue

			
			rqst_id = raw[0]
			#cmd = raw[1] #we don't use it
			state = raw[2]
			
			self._ack_lock.acquire()
			
			try:
				if state not in ['ACK', 'TERM']:
					raise PocoRemoteError("Unknown state " + state +"!")
					
				self.requests[rqst_id].state = state
				
				if state == 'ACK':
					logging.debug("Got a ACK for rqst " + rqst_id)
					self.requests[rqst_id].activity_id = raw[3]
					
				if state == 'TERM':
					logging.debug("Got a TERM for rqst " + rqst_id)
					
					if raw[3] == 'OK':
						self.requests[rqst_id].status = 'OK'
						self.requests[rqst_id].answer = raw[4:]
						
						if self.requests[rqst_id].callback: #invoke the callback
							self.requests[rqst_id].callback(('OK', raw[4:]))
							#Release the request id
							self._rpc("ACK " + rqst_id)
						
					else:
						self.requests[rqst_id].status = 'FAILED'
						self.requests[rqst_id].answer = raw[3] #store the error message in Request.answer
						
						if self.requests[rqst_id].callback: #invoke the callback
							self.requests[rqst_id].callback(('FAILED', raw[3]))
							#TODO: Callbacks should not be launch from this thread. Spawn a thread?
							#Release the request id
							self._rpc("ACK " + rqst_id)
			
			except KeyError:
				raise PocoRemoteError("Got a request ID I don't know!")
			finally:
				self._ack_lock.release()

	def _rpc(self, rqst):
		"""Executes a simple blocking RPC call to TCL serv, checking the status
		of the answer (OK or ERROR)
		"""

		logging.debug(">>> " + rqst)
		
		self._tclserv.write(rqst + "\r\n")
		self._tclserv.flush()
		
		res = self._poco_responses_queue.get()
		
		if res[0] != "OK":
			raise PocoRemoteError(" ".join(res[2:]))
		
		return res[1:]
	
	def _read_tclserv_answer(self):
		try:
			raw = self._tclserv.readline().splitlines()[0].split()
		except IndexError:
			return None
		return raw
	
	def load(self, module_name, module_path = None, start_module = True):
		"""Loads a module on a remote server.
		
		If the flag start_module is set to True (default), the server will 
		try to (re)start the module *if a first connection attempt fails*.
		"""
		
		if type(module_name) == type(list()):
			return [self.load(module, start_module) for module in module_name]

		logging.debug("Loading module " + module_name)
		
		try:
			self._rpc("LM " + module_name + ' ' + (module_path or ''))
		except PocoRemoteError as pre:
			raise PocoRemoteError("Error while loading module " + module_name + ": " + pre.value)
		
		try:
			self._rpc("cs::init")
		except PocoRemoteError as pre:
			if 'too small' in pre.value:
				logging.debug("Server mailbox too small. Rebuilding it.")
				try:
					self._rpc("modules::disconnect")
					self._rpc("cs::exit")
					self._rpc("cs::init")
				except PocoRemoteError as pre:
					raise PocoRemoteError("Error while re-initializing TCL serv with new module: " + pre.value)
					
		try:
			self._rpc("modules::connect")
			logging.info("Module " + module_name + " successfully loaded.")
		except PocoRemoteError as pre:
			logging.debug("First attempt to connect to module " + module_name + \
			" failed. Trying to start it on the remote server...")
			try:
				self.execute((module_path or '') + module_name + ' -b')
			except PocoRemoteError as pre:
				raise PocoRemoteError("Error while trying to start module " + module_name + ": " + pre.value)
			
			logging.debug("Module " + module_name + " started.")
			
			try:
				self._rpc("modules::connect")
				logging.info("Module " + module_name + " successfully loaded.")
			except PocoRemoteError as pre:
				raise PocoRemoteError("Error while connecting to module " + module_name + ": " + pre.value)
		
		module = PocoModule(module_name, self)
		self._loaded_modules[module_name] = module
		
		return module
	
	def autoload(self):
		""" Tries to load all available modules on the remote server, and return
		a list of corresponding PocoModules.
		"""
		
		res = []
		
		for module in self.modules():
			try:
				res.append(self.load(module))
			except PocoRemoteError as pre:
				if 'no files matched glob pattern' in pre.value:
					logging.log(4, "TCLserv couldn't find TCL client for module " + module + ". I skip it.")
				else:
					raise
		
		return res
		
	def unload(self, module_name, kill_module = False):
		"""Unload a module from the remote server.
		
		If the flag kill_module is set to True, the server will try to kill as 
		well the module process.
		"""
		
		if type(module_name) == type(list()):
			for module in module_name:
				self.unload(module, kill_module)
			
		rqst = "UNLM " + module_name
		
		try:
			self._rpc(rqst)
		except PocoRemoteError as pre:
			raise PocoRemoteError("Error while unloading module " + module_name + ": " + pre.value)
		
		if kill_module:
			rqst = "KILL " + module_name + " abort"
			
			try:
				self._rpc(rqst)
			except PocoRemoteError as pre:
				raise PocoRemoteError("Error while killing module " + module_name + ": " + pre.value)
		
		self._loaded_modules.pop(module_name)
		
		logging.info("Module " + module_name + " successfully unloaded")
		
	def reload(self, module_name, restart_module = True):
		"""Reloads a module on the remote server.
		If restart_module is set to True (default), the process corresponding to
		the module is restart as well. Else, only the TCL bindings are reloaded.
		"""
		
		if type(module_name) == type(list()):
			return [self.reload(module, restart_module) for module in module_name]
				
		try:
			self.unload(module_name, restart_module)
		except PocoRemoteError as pre:
			if 'No such module' in pre.value:
				pass
			else:
				raise
			
		return self.load(module_name, None, restart_module)
		
	def request(self, request, callback = None, wait_for_ack = True):
		"""Send a request to a remote server.
		* If callback is None, the call is blocking and will return the 
		request result. Otherwise, a request id is answered that can be used to 
		check later on the request status (with PocoServProxy.status()) or abort it. The 
		callback method is besides invoked when the request finishes.
		* If wait_for_ack is set to true, the call only returns when the server
		has ackwoledged the request. Note: for control requests, the server does
		not send acknwoledgements but directly return the request result. For 
		these requests, the wait_for_ack flag is ignored.
		
		TODO: doc callback
		"""
		
		rqst = "RQST " + request
		
		rqst_id = ""
		self._ack_lock.acquire()
		try:
			[rqst_id] = self._rpc(rqst)
			
			self.requests[rqst_id] = PocoRequest(rqst_id) # instanciate a new pending request
			self.requests[rqst_id].callback = callback
		
		except PocoRemoteError as pre:
			raise PocoRemoteError("Error while sending a request: " + pre.value)
		finally:
			self._ack_lock.release()
		
		if not callback:
			while self.requests[rqst_id].state != 'TERM': #block until we get an answer
				time.sleep(0.03) #sleep 30ms
				
			#Release the request id
			self._rpc("ACK " + rqst_id)
			
			#TODO: should we purge _pending_request once they are done?
			return (self.requests[rqst_id].status, self.requests[rqst_id].answer)
		
		else:
			if wait_for_ack:
				while self.requests[rqst_id].state not in ['ACK', 'TERM']: #block until we get an ack (or an answer)
					time.sleep(0.03) #sleep 30ms
				
			return rqst_id # returns the request id
		
	def abort(self, id):
		"""Try to abort a remote activity.
		
		Returns True is the task has been successfully aborted, False if the
		server couldn't interrupt the task.
		"""
		
		if id in self.requests.keys():
			#Our activity_id is a request id! caramba!
			#try to stop all activities
			self.abort(self.requests[id].activity_id)
			
		rqst = "ABORT " + id
		
		try:
			return self._rpc(rqst)
		except PocoRemoteError as pre:
			raise PocoRemoteError("Error while aborting request: " + pre.value)
		
		
	def modules(self):
		"""Returns the list of modules currently running on the remote host.
		
		Note: the list is actually the list of Pocolibs mailboxes. For Genom 
		modules, this matches the module name, but a specific application may
		have more than one mailbox.
		"""
		
		rqst = "cs::lsmbox"
		
		try:
			return self._rpc(rqst)
		except PocoRemoteError as pre:
			raise PocoRemoteError("Error while retrieving the list of modules: " + pre.value)

	def execute(self, cmd):
		"""Executes arbitrary TCL code on the server. Use with caution!
		"""
		
		rqst = "exec -- " + cmd + (' &' if not cmd.endswith(' &') else '')
		
		try:
			return self._rpc(rqst)
		except PocoRemoteError as pre:
			raise PocoRemoteError("Error while retrieving the list of modules: " + pre.value)
	
	def list_services(self, module_name):
		"""Returns the list of requests exposed by a module
		"""
		
		rqst = "info commands " + module_name + "::*Send"
		
		raw = []
		try:
			raw = self._rpc(rqst)
		except PocoRemoteError as pre:
			raise PocoRemoteError("Error while retrieving the list of requests for module " + module_name + ": " + pre.value)
		
		return [rqst.split('::')[2][:-4] for rqst in raw]
	
	def close(self):
		self._running = False
		
		self._tclserv.write("BYE\r\n")
		self._tclserv.flush()
		
		self.join()
		logging.log(logging.DEBUG, 'Closing the connection to TCL serv...')
		self._tclserv.close()
		self.s.close()
		
		logging.log(logging.DEBUG, 'Done. Bye bye!')
	
	def pending_requests(self):
		"""Returns the list of all request ids that are not finished yet.
		"""
		return [x for x in requests if x.state != 'TERM']
	
	def __del__(self):
		if self._tclserv:
			self.close()

def discover(hosts = 'localhost', port = 9472):
	"""Connects to all servers listed in the hosts argument. On each of them,
	loads all available modules.
	
	hosts can be a list of hostnames that all use the same port (as specified by
	the port argument (default: 9472), a list of tuples (hostname, port), or a 
	mix of both.
	
	The method returns a tuple (dict of servers, dict of all modules).
	"""
	
	if isinstance(hosts, basestring) or isinstance(hosts, tuple):
		hosts = [hosts]
		
	servers = {}
	modules = {}
	
	for host in hosts:
		if not isinstance(host, tuple):
			host = (host, port)
		
		#check if the host already exist in the server list (in case of several
		#TCL serv on different port). Is this case, append the port to the host name.
		key = host[0]
		if key in servers:
			key += ':' + str(host[1])
		
		servers[key] = PocoServProxy(host)
	
	for server in servers:
		servers[server].autoload()
		modules.update(servers[server]._loaded_modules)
	
	logging.info("Loaded modules " + ", ".join(modules) + " on " + str(len(servers)) + \
	" server(s).")
	
	return (servers, modules)
	

if __name__ == '__main__':

	logging.basicConfig(level=logging.DEBUG, format="%(message)s")

	def printer(c):
		print("Yeahh! request content: " + str(c))
	
	print("Starting now...")
	try:
		servers, modules = discover()
		print("Available modules: " + str(modules))
		
		pom = modules['pom']
		
		print (pom.GetModel())
		
		print ("Finished!")
		
		for s in servers:
			servers[s].close()

	except PocoRemoteError as ose:
		print('Oups! An error occured!')
		print(ose)
	finally:
		pass
